/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 * 
 * Author: Siamak Haschemi
 * Contact: haschemi@informatik.hu-berlin.de
 */
package net.sourceforge.osgi.deployment.maven;

import java.io.File;
import java.util.List;

import net.sourceforge.osgi.deployment.maven.container.DeploymentPackageInfo;

import org.apache.maven.artifact.factory.ArtifactFactory;
import org.apache.maven.artifact.handler.manager.ArtifactHandlerManager;
import org.apache.maven.artifact.repository.ArtifactRepository;
import org.apache.maven.artifact.resolver.ArtifactResolver;
import org.apache.maven.plugin.logging.Log;
import org.apache.maven.project.MavenProject;
import org.codehaus.plexus.archiver.manager.ArchiverManager;


/**
 * The deployment context provides all relevant data to the deployment plugin. Currently it depends on maven classes and interfaces, but the idea is to abstract
 * from maven in order to use the deployment plugin with other technologies.
 * 
 * @author Siamak Haschemi, haschemi@informatik.hu-berlin.de
 * 
 */
public interface IDeploymentPluginContext {

  /**
   * @return the Maven logger
   */
  Log getLogger();

  /**
   * @return the outputDirectory
   */
  File getOutputDirectory();

  /**
   * @return the baseDir
   */
  File getBaseDir();

  /**
   * @return the manifestLocation
   */
  File getManifestLocation();

  /**
   * @return the project
   */
  MavenProject getProject();

  /**
   * @return the buildDirectory
   */
  String getBuildDirectory();

  /**
   * @return the supportedProjectTypes
   */
  List< String > getSupportedProjectTypes();

  /**
   * @return the localRepository
   */
  ArtifactRepository getLocalRepository();

  /**
   * @return the remoteRepositories
   */
  List< ArtifactRepository > getRemoteRepositories();

  /**
   * @return the artifactFactory
   */
  ArtifactFactory getArtifactFactory();

  /**
   * @return the artifactResolver
   */
  ArtifactResolver getArtifactResolver();

  /**
   * @return the artifactHandlerManager
   */
  ArtifactHandlerManager getArtifactHandlerManager();

  /**
   * @return the archiverManager
   */
  ArchiverManager getArchiverManager();

  /**
   * @return the deploymentPackage
   */
  DeploymentPackageInfo getDeploymentPackageInfo();

  /**
   * @return the writeExtraData
   */
  boolean isWriteExtraData();

  /**
   * Returns the name of the plugin.
   * 
   * @return the name of the plugin
   */
  String getPluginName();

  /**
   * Returns the plugin version.
   * 
   * @return the plugin version
   */
  String getPluginVersion();

  /**
   * This method resolves an artifact on all available repositories and returns the file handle to that artifact.
   * 
   * @param p_groupId
   *          the groupId of the artifact to resolve
   * @param p_artifactId
   *          the artifactId of the artifact to resolve
   * @param p_version
   *          the version of the artifact to resolve
   * @return the resolved file handle of the artifact
   */
  File resolveResource(final String p_groupId, final String p_artifactId, final String p_version);

}
