/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 * 
 * Author: Siamak Haschemi
 * Contact: haschemi@informatik.hu-berlin.de
 */
package net.sourceforge.osgi.deployment.maven.manifest;

import java.io.File;
import java.io.FileInputStream;
import java.util.Collection;
import java.util.List;
import java.util.jar.JarInputStream;
import java.util.jar.Manifest;
import java.util.regex.Pattern;

import net.sourceforge.osgi.deployment.maven.IDeploymentPluginContext;
import net.sourceforge.osgi.deployment.maven.container.BundleResource;
import net.sourceforge.osgi.deployment.maven.container.DeploymentPackageInfo;
import net.sourceforge.osgi.deployment.maven.container.ProcessedResource;

import org.apache.maven.model.License;
import org.apache.maven.project.MavenProject;


/**
 * This class represents the manifest of a Deployment-Package. It creates the various headers using the {@link IDeploymentPluginContext}. The
 * Deployment-Package manifest is an ordered manifest, because the order and empty lines are part of the semantics.
 * 
 * @author Siamak Haschemi, haschemi@informatik.hu-berlin.de
 */
public class DeploymentPackageManifest extends OrderedManifest {
  /** Extra header in the manifest file. */
  public static final String CREATED_BY = "Created-By";
  /** Extra header in the manifest file. */
  public static final String TOOL = "Tool";
  /** Extra header in the manifest file. */
  public static final String CREATED_AT = "Created-At";

  /** Constant from the Deployment Admin Specification in the OSGi Platform Release 4, Version 4.1. */
  public static final String DP_MANIFESTVERSION = "DeploymentPackage-ManifestVersion";

  /** Constant from the Deployment Admin Specification in the OSGi Platform Release 4, Version 4.1. */
  public static final String DP_CONTENT_TYPE = "application/vnd.osgi.dp";

  /** Constant from the Deployment Admin Specification in the OSGi Platform Release 4, Version 4.1. */
  public static final String DP_SYMBOLICNAME = "DeploymentPackage-SymbolicName";

  /** Constant from the Deployment Admin Specification in the OSGi Platform Release 4, Version 4.1. */
  public static final String DP_VERSION = "DeploymentPackage-Version";

  /** Constant from the Deployment Admin Specification in the OSGi Platform Release 4, Version 4.1. */
  public static final String DP_COPYRIGHT = "DeploymentPackage-Copyright";

  /** Constant from the Deployment Admin Specification in the OSGi Platform Release 4, Version 4.1. */
  public static final String DP_CONTACTADDRESS = "DeploymentPackage-ContactAddress";

  /** Constant from the Deployment Admin Specification in the OSGi Platform Release 4, Version 4.1. */
  public static final String DP_DESCRIPTION = "DeploymentPackage-Description";

  /** Constant from the Deployment Admin Specification in the OSGi Platform Release 4, Version 4.1. */
  public static final String DP_DOCURL = "DeploymentPackage-DocURL";

  /** Constant from the Deployment Admin Specification in the OSGi Platform Release 4, Version 4.1. */
  public static final String DP_VENDOR = "DeploymentPackage-Vendor";

  /** Constant from the Deployment Admin Specification in the OSGi Platform Release 4, Version 4.1. */
  public static final String DP_LICENSE = "DeploymentPackage-License";

  /** Constant from the Deployment Admin Specification in the OSGi Platform Release 4, Version 4.1. */
  public static final String DP_FIXPACK = "DeploymentPackage-FixPack";

  /** Constant from the Deployment Admin Specification in the OSGi Platform Release 4, Version 4.1. */
  public static final String DP_CUSTOMIZER = "DeploymentPackage-Customizer";

  /** Constant from the Deployment Admin Specification in the OSGi Platform Release 4, Version 4.1. */
  public static final String DP_NAME = "Name";

  /** Constant from the Deployment Admin Specification in the OSGi Platform Release 4, Version 4.1. */
  public static final String DP_SHA1_DIGEST = "SHA1-Digest";

  /** Constant from the Deployment Admin Specification in the OSGi Platform Release 4, Version 4.1. */
  public static final String DP_RESOURCE_PROCESSOR = "Resource-Processor";

  /** Constant from the Deployment Admin Specification in the OSGi Platform Release 4, Version 4.1. */
  public static final Pattern PATH_NAME_PATTERN = Pattern.compile("[A-Za-z0-9_\\.-]+");

  /** Constant from the OSGi Platform Release 4, Version 4.1. */
  public static final Pattern UNIQUE_NAME_PATTERN = Pattern.compile(".+(\\..+)*");

  /**
   * Manifest header (named &quot;Bundle-Version&quot;) identifying the bundle's version.
   */
  public static final String BUNDLE_VERSION = "Bundle-Version";

  /**
   * Manifest header (named &quot;Bundle-SymbolicName&quot;) identifying the bundle's symbolic name.
   */
  public static final String BUNDLE_SYMBOLICNAME = "Bundle-SymbolicName";

  /**
   * Manifest header (named &quot;Bundle-Localization&quot;) identifying the base name of the bundle's localization entries.
   */
  public static final String BUNDLE_LOCALIZATION = "Bundle-Localization";

  /**
   * @param p_context
   *          the context
   */
  public DeploymentPackageManifest(final IDeploymentPluginContext p_context) {

    try {
      writeGlobalSection(p_context);
      writeNameSection(p_context);
    } catch (final ManifestException e) {
      throw new ManifestException("The creation of the Deploymant-Package Manifest file failed", e);
    }
  }

  private void writeNameSection(final IDeploymentPluginContext p_context) {
    writeBundleResources(p_context);
    writeLocalizationResources(p_context);
    writeProcessedResources(p_context);
  }

  private void writeGlobalSection(final IDeploymentPluginContext p_context) {
    // The manifest version of the overall jar manifest
    add(MANIFEST_VERSION, "1.0");

    writeExtraData(p_context);

    // The deployment version in the 4.1 Spec.
    add(DP_MANIFESTVERSION, "1");
    add(CONTENT_TYPE, DP_CONTENT_TYPE);

    final DeploymentPackageInfo deploymentPackageInfo = p_context.getDeploymentPackageInfo();
    final MavenProject project = p_context.getProject();

    writesymlobicName(deploymentPackageInfo, project);
    writeVersion(deploymentPackageInfo, project);
    writeFixPack(deploymentPackageInfo);
    writeDescription(deploymentPackageInfo, project);
    writeLicense(project);
    writeVendor(deploymentPackageInfo, project);
    writeCopyright(deploymentPackageInfo);
    writeContactAdress(deploymentPackageInfo);
    writeDocURL(deploymentPackageInfo);
    writeBundleLocalization(deploymentPackageInfo);
  }

  private void writeProcessedResources(final IDeploymentPluginContext p_context) {
    final DeploymentPackageInfo deploymentPackageInfo = p_context.getDeploymentPackageInfo();

    for (final ProcessedResource processedResource : deploymentPackageInfo.getProcessedResources()) {
      final String resourceId = processedResource.getResourceId();

      if (!isResourceId(resourceId)) {
        throw new ManifestException("\"" + resourceId + "\" is not a valid path name as specified in OSGi 114.3.2 Service Compendium");
      }

      add(OrderedManifest.EMPTY_LINE);
      add(DP_NAME, resourceId);
      add(DP_RESOURCE_PROCESSOR, processedResource.getProcessor());
    }
  }

  private void writeLocalizationResources(final IDeploymentPluginContext p_context) {
    final DeploymentPackageInfo deploymentPackageInfo = p_context.getDeploymentPackageInfo();

    for (final ProcessedResource processedResource : deploymentPackageInfo.getLocalizationResources()) {
      final String resourceId = processedResource.getResourceId();

      if (!isResourceId(resourceId)) {
        throw new ManifestException("\"" + resourceId + "\" is not a valid path name as specified in OSGi 114.3.2 Service Compendium");
      }

      add(OrderedManifest.EMPTY_LINE);
      add(DP_NAME, resourceId);
      add(DP_RESOURCE_PROCESSOR, processedResource.getProcessor());
    }
  }

  private void writeBundleResources(final IDeploymentPluginContext p_context) {
    final DeploymentPackageInfo deploymentPackageInfo = p_context.getDeploymentPackageInfo();

    for (final BundleResource bundleResource : deploymentPackageInfo.getBundleResources()) {
      try {

        // Get some info from the bundle jar and add them to the manifest
        final File resolvedFile = bundleResource.getResolvedFile();
        final JarInputStream jis = new JarInputStream(new FileInputStream(resolvedFile));
        final Manifest bundleManifest = jis.getManifest();
        final String bundleVersion = bundleManifest.getMainAttributes().getValue(BUNDLE_VERSION);
        final String bundleSymbolicName = bundleManifest.getMainAttributes().getValue(BUNDLE_SYMBOLICNAME);
        final String resourceId = bundleResource.getResourceId();

        if (!isUniqueName(bundleSymbolicName)) {
          throw new ManifestException("\"" + bundleSymbolicName + "\" is not a valid unique-name as specified in OSGi 1.3.2 Core");
        }

        if (!isVersion(bundleVersion)) {
          throw new ManifestException("\"" + bundleVersion + "\" is not a valid version as specified in OSGi 3.2.4 Core");
        }

        if (!isResourceId(resourceId)) {
          throw new ManifestException("\"" + resourceId + "\" is not a valid path name as specified in OSGi 114.3.2 Service Compendium");
        }

        add(OrderedManifest.EMPTY_LINE);
        add(DP_NAME, resourceId);
        if (bundleResource.isCustomizer()) {
          add(DP_CUSTOMIZER, "true");
        }
        add(BUNDLE_SYMBOLICNAME, bundleSymbolicName);
        add(BUNDLE_VERSION, bundleVersion);
      } catch (final Exception e) {
        throw new ManifestException("Error while processing bundle resource " + bundleResource, e);
      }
    }
  }

  private void writeBundleLocalization(final DeploymentPackageInfo p_deploymentPackageInfo) {
    // Write the Bundle-Localization
    if (isNotEmpty(p_deploymentPackageInfo.getBundleLocalization())) {
      add(BUNDLE_LOCALIZATION, p_deploymentPackageInfo.getBundleLocalization());
    }
  }

  private void writeDocURL(final DeploymentPackageInfo p_deploymentPackageInfo) {
    // Write the DeploymentPackage-DocURL
    if (isNotEmpty(p_deploymentPackageInfo.getDocURL())) {
      add(DP_DOCURL, p_deploymentPackageInfo.getDocURL());
    }
  }

  private void writeContactAdress(final DeploymentPackageInfo p_deploymentPackageInfo) {
    // Write the DeploymentPackage-ContactAdress
    if (isNotEmpty(p_deploymentPackageInfo.getContactAddress())) {
      add(DP_CONTACTADDRESS, p_deploymentPackageInfo.getContactAddress());
    }
  }

  private void writeCopyright(final DeploymentPackageInfo p_deploymentPackageInfo) {
    // Write the DeploymentPackage-Copyright
    if (isNotEmpty(p_deploymentPackageInfo.getCopyright())) {
      add(DP_COPYRIGHT, p_deploymentPackageInfo.getCopyright());
    }
  }

  private void writeVendor(final DeploymentPackageInfo p_deploymentPackageInfo, final MavenProject p_project) {
    // Write the DeploymentPackage-Vendor
    String vendor = null;
    if (isNotEmpty(p_deploymentPackageInfo.getVendor())) {
      vendor = p_deploymentPackageInfo.getVendor();
    } else if (p_project.getOrganization() != null) {
      vendor = p_project.getOrganization().getName();
    }
    if (isNotEmpty(vendor)) {
      header(DP_VENDOR, vendor);
    }
  }

  @SuppressWarnings("unchecked")
  private void writeLicense(final MavenProject p_project) {
    // Write the DeploymentPackage-License
    final String licenseText = printLicenses(p_project.getLicenses());
    if (isNotEmpty(licenseText)) {
      header(DP_LICENSE, licenseText);
    }
  }

  private void writeDescription(final DeploymentPackageInfo p_deploymentPackageInfo, final MavenProject p_project) {
    // Write the DeploymentPackage-Description
    String description = null;
    if (isNotEmpty(p_deploymentPackageInfo.getDescription())) {
      description = p_deploymentPackageInfo.getDescription();
    } else {
      description = p_project.getDescription();
    }
    if (description != null) {
      header(DP_DESCRIPTION, description);
    }
  }

  private void writeFixPack(final DeploymentPackageInfo p_deploymentPackageInfo) {
    final String fixPack = p_deploymentPackageInfo.getFixPack();
    if (isNotEmpty(fixPack)) {
      if (!isVersionRange(fixPack)) {
        throw new ManifestException("\"" + fixPack + "\" is not a valid version range as specified in OSGi 3.2.5 Core");
      }
      add(DP_FIXPACK, fixPack);
    }
  }

  private void writeVersion(final DeploymentPackageInfo p_deploymentPackageInfo, final MavenProject p_project) {
    // Write the DeploymentPackage-Version
    String version = null;
    if (isNotEmpty(p_deploymentPackageInfo.getVersion())) {
      version = p_deploymentPackageInfo.getVersion();
    } else {
      String pomVersion = p_project.getVersion();
      // Make the POM version OSGi conform
      if (!isVersion(pomVersion)) {
        pomVersion = pomVersion.replaceAll("-", "\\.");
      }
      version = pomVersion;
    }
    if (!isVersion(version)) {
      throw new ManifestException("\"" + version + "\" is not a valid version as specified in OSGi 3.2.4 Core");
    }
    add(DP_VERSION, version);
  }

  private void writesymlobicName(final DeploymentPackageInfo p_deploymentPackageInfo, final MavenProject p_project) {
    // Write the DeploymentPackage-SymbolicName
    String symbolicName = null;
    if (p_deploymentPackageInfo.getSymbolicName() != null && p_deploymentPackageInfo.getSymbolicName().trim().length() > 0) {
      symbolicName = p_deploymentPackageInfo.getSymbolicName();
    } else {
      symbolicName = p_project.getGroupId() + "." + p_project.getArtifactId();
    }
    if (!isUniqueName(symbolicName)) {
      throw new ManifestException("\"" + symbolicName + "\" is not a valid unique-name as specified in OSGi 1.3.2 Core");
    }
    add(DP_SYMBOLICNAME, symbolicName);
  }

  private void writeExtraData(final IDeploymentPluginContext p_context) {
    // Write some extra data
    if (p_context.isWriteExtraData()) {
      add(CREATED_BY, System.getProperty("java.version") + " (" + System.getProperty("java.vendor") + ")");
      add(TOOL, p_context.getPluginName() + " " + p_context.getPluginVersion());
      add(CREATED_AT, "" + System.currentTimeMillis());
    }
  }

  private void header(final String p_key, final Object p_value) {
    if (p_value == null) {
      return;
    }

    if (p_value instanceof Collection && ((Collection< ? >) p_value).isEmpty()) {
      return;
    }

    add(p_key, p_value.toString().replaceAll("[\r\n]", ""));
  }

  private String printLicenses(final List< License > p_licenses) {
    if (p_licenses == null || p_licenses.size() == 0) {
      return null;
    }
    final StringBuffer sb = new StringBuffer();
    String del = "";
    for (final License license : p_licenses) {
      final String url = license.getUrl();
      if (url == null) {
        continue;
      }
      sb.append(del);
      sb.append(url);
      del = ", ";
    }
    if (sb.length() == 0) {
      return null;
    }
    return sb.toString();
  }

  private static boolean isNotEmpty(final String p_string) {
    if (p_string != null && p_string.trim().length() > 0) {
      return true;
    }
    return false;
  }

  /**
   * This method checks whether the given version conforms to the OSGi Specification.
   * 
   * @param p_versionToCheck
   *          the version to check
   * @return <CODE>TRUE></CODE> if the version is valid, else <CODE>FALSE</CODE>
   */
  public static boolean isVersion(final String p_versionToCheck) {
    // TODO implement version check
    return true;
  }

  /**
   * This method checks whether the given version range conforms to the OSGi Specification.
   * 
   * @param p_versionRange
   *          the version range to check
   * @return <CODE>TRUE></CODE> if the version range is valid, else <CODE>FALSE</CODE>
   */
  public static final boolean isVersionRange(final String p_versionRange) {
    // TODO implement versionRange check
    return true;
  }

  /**
   * This method checks whether the given symbolic name conforms to the OSGi Specification of a unique name.
   * 
   * @param p_symbolicName
   *          the symbolic name to check
   * @return <CODE>TRUE></CODE> if the symbolic name is valid, else <CODE>FALSE</CODE>
   */
  public static boolean isUniqueName(final String p_symbolicName) {
    boolean isUniqueName = true;
    // DeploymentPackage-SymbolicName ::= unique-name
    // unique-name ::= identifier ( '.' identifier )*
    // identifier ::= jletter jletterordigit * jletter ::= <see [5] Lexical Structure Java Language for JavaLetter>
    // jletterordigit::= <See [5] Lexical Structure Java Language for JavaLetterOrDigit >

    for (int i = 0; i < p_symbolicName.length(); i++) {
      final char c = p_symbolicName.charAt(i);
      if (!(Character.isJavaIdentifierStart(c) || Character.isJavaIdentifierPart(c) || c == '.')) {
        isUniqueName = false;
      }
    }

    if (isUniqueName && !UNIQUE_NAME_PATTERN.matcher(p_symbolicName).matches()) {
      isUniqueName = false;
    }

    return isUniqueName;
  }

  /**
   * This method checks whether the given resourceId conforms to the OSGi Specification.
   * 
   * @param p_resourceId
   *          the resourceId to check
   * @return <CODE>TRUE></CODE> if the resourceId is valid, else <CODE>FALSE</CODE>
   */
  public static boolean isResourceId(final String p_resourceId) {
    if (p_resourceId == null) {
      return false;
    }
    if (p_resourceId.contains("/")) {

      for (final String pathToken : p_resourceId.split("/")) {
        if (!pathToken.equals("") && !PATH_NAME_PATTERN.matcher(pathToken).matches()) {
          return false;
        }
      }
    } else {
      if (!PATH_NAME_PATTERN.matcher(p_resourceId).matches()) {
        return false;
      }
    }
    return true;
  }

}
