/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 * 
 * Author: Siamak Haschemi
 * Contact: haschemi@informatik.hu-berlin.de
 */
package net.sourceforge.osgi.deployment.maven.manifest;

import java.io.DataOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.regex.Pattern;

/**
 * The {@link OrderedManifest} stores manifest entries in a defined order.
 * 
 * @author Siamak Haschemi, haschemi@informatik.hu-berlin.de
 */
public class OrderedManifest {
  /**
   * <code>Manifest-Version</code> manifest attribute. This attribute indicates the version number of the manifest standard to which a JAR file's manifest
   * conforms.
   * 
   * @see <a href="http://java.sun.com/j2se/1.4.2/docs/guide/jar/jar.html#JAR%20Manifest"> Manifest and Signature Specification</a>
   */
  public static final String MANIFEST_VERSION = "Manifest-Version";
  
  /**
   * A pattern to verify the header of a manifest entry.
   */
  public static final Pattern HEADER_PATTERN = Pattern.compile("[A-Za-z0-9][-a-zA-Z0-9_]+");

  /**
   * <code>Content-Type</code> manifest attribute.
   */
  protected static final String CONTENT_TYPE = "Content-Type";

  /**
   * An empty line on the manifest.
   */
  protected static final ManifestEntry EMPTY_LINE = new ManifestEntry("", "");

  private static final int MAXIMUM_MANIFEST_LINE_WIDTH = 72;
  private final List< ManifestEntry > m_entries = Collections.synchronizedList(new ArrayList< ManifestEntry >());

  /**
   * This class represents an entry in the {@link OrderedManifest}.
   * 
   * @author Siamak Haschemi, haschemi@informatik.hu-berlin.de
   */
  private static class ManifestEntry {
    private final String m_name;
    private final String m_value;

    public ManifestEntry(final String p_name, final String p_value) {
      if (p_name == null || p_value == null) {
        throw new IllegalArgumentException("NULL is not allowed as parameter");
      }

      final boolean nameEmpty = p_name.trim().length() == 0;
      final boolean nameNotEmpty = p_name.trim().length() > 0;
      final boolean valueNotEmpty = p_value.trim().length() > 0;

      if (nameEmpty && valueNotEmpty) {
        throw new IllegalArgumentException("name param was empty, so value param has also to be.");
      }

      if (nameNotEmpty && !HEADER_PATTERN.matcher(p_name).matches()) {
        throw new IllegalArgumentException("Header name not matches OSGi specification.");
      }
      m_name = p_name;
      m_value = p_value;
    }

    @Override
    public boolean equals(final Object p_obj) {
      if (p_obj == null) {
        return false;
      }

      if (!this.getClass().equals(p_obj.getClass())) {
        return false;
      }

      final ManifestEntry obj = (ManifestEntry) p_obj;
      if (m_name.equals(obj.m_name) && m_value.equals(obj.m_value)) {
        return true;
      }
      return false;
    }

    @Override
    public int hashCode() {
      return m_name.hashCode() + m_value.hashCode();
    }
  };

  /**
   * Add a {@link ManifestEntry} to the {@link OrderedManifest}.
   * 
   * @param p_manifestEntry
   *          the {@link ManifestEntry} to add
   */
  public final void add(final ManifestEntry p_manifestEntry) {
    m_entries.add(p_manifestEntry);
  }

  /**
   * Add a key/value pair to the {@link OrderedManifest}.
   * 
   * @param p_header
   *          The header to add
   * @param p_value
   *          the value for the header to add
   */
  public final void add(final String p_header, final String p_value) {
    m_entries.add(new ManifestEntry(p_header, p_value));
  }

  /**
   * Write the current contents of the {@link OrderedManifest} into a {@link OutputStream}.
   * 
   * @param p_out
   *          the {@link OutputStream} to write into
   * @throws IOException
   *           If the writing of the {@link OrderedManifest} fails.
   */
  public final void write(final OutputStream p_out) throws IOException {
    final DataOutputStream dos = new DataOutputStream(p_out);

    for (final ManifestEntry manifestEntry : m_entries) {
      if (manifestEntry.equals(EMPTY_LINE)) {
        dos.writeBytes("\r\n");
        continue;
      }

      final StringBuffer buffer = new StringBuffer(manifestEntry.m_name);
      buffer.append(": ");

      String value = manifestEntry.m_value;
      final byte[] vb = value.getBytes("UTF8");
      // value = new String(vb, 0, 0, vb.length);
      value = new String(vb, "UTF8");
      buffer.append(value);

      buffer.append("\r\n");
      OrderedManifest.make72Safe(buffer);
      dos.writeBytes(buffer.toString());
    }
    dos.writeBytes("\r\n");
    dos.flush();
  }

  /**
   * <p>
   * From the manifest specification:
   * </p>
   * <p>
   * <i>"No line may be longer than 72 bytes (not characters), in its UTF8-encoded form. If a value would make the initial line longer than this, it should be
   * continued on extra lines (each starting with a single SPACE). "</i> <a
   * href="http://java.sun.com/j2se/1.4.2/docs/guide/jar/jar.html">http://java.sun.com/j2se/1.4.2/docs/guide/jar/jar.html</a>
   * </p>
   * Adds line breaks to enforce a maximum 72 bytes per line.
   * 
   * @param p_line
   *          the line to make safe
   */
  static void make72Safe(final StringBuffer p_line) {
    final String lineBreak = "\r\n";
    final String space = " ";
    final String lineBreakAndSpace = lineBreak + space;

    int length = p_line.length();
    if (length > MAXIMUM_MANIFEST_LINE_WIDTH) {
      int index = MAXIMUM_MANIFEST_LINE_WIDTH - lineBreak.length();
      while (index < length - 2) {
        p_line.insert(index, lineBreakAndSpace);
        index += MAXIMUM_MANIFEST_LINE_WIDTH;
        length += lineBreakAndSpace.length();
      }
    }
    return;
  }
}
