/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 * 
 * Author: Siamak Haschemi
 * Contact: haschemi@informatik.hu-berlin.de
 */
package net.sourceforge.osgi.deployment.maven.mojo;

import java.io.File;
import java.util.Arrays;
import java.util.List;

import net.sourceforge.osgi.deployment.maven.DeploymentPluginException;
import net.sourceforge.osgi.deployment.maven.DeploymentPlugin;
import net.sourceforge.osgi.deployment.maven.IDeploymentPluginContext;
import net.sourceforge.osgi.deployment.maven.container.BundleResource;
import net.sourceforge.osgi.deployment.maven.container.DeploymentPackageInfo;

import org.apache.maven.artifact.Artifact;
import org.apache.maven.artifact.factory.ArtifactFactory;
import org.apache.maven.artifact.handler.manager.ArtifactHandlerManager;
import org.apache.maven.artifact.repository.ArtifactRepository;
import org.apache.maven.artifact.resolver.ArtifactResolver;
import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.plugin.logging.Log;
import org.apache.maven.project.MavenProject;
import org.codehaus.plexus.archiver.manager.ArchiverManager;


/**
 * Create an OSGi deployment package from Maven project.
 * 
 * @goal deployment
 * @phase package
 * @requiresDependencyResolution runtime
 * @description build an OSGi deployment package jar
 * @author Siamak Haschemi, haschemi@informatik.hu-berlin.de
 */
public class DeploymentPluginMojoContext extends AbstractMojo implements IDeploymentPluginContext {
  /**
   * The directory for the generated bundles.
   * 
   * @parameter expression="${project.build.outputDirectory}"
   * @required
   */
  private File m_outputDirectory;

  /**
   * The directory for the pom.
   * 
   * @parameter expression="${basedir}"
   * @required
   */
  private File m_baseDir;

  /**
   * @parameter property="deploymentPackage"
   */
  private DeploymentPackageInfo m_deploymentPackageInfo = null;

  /**
   * Directory where the manifest will be written.
   * 
   * @parameter expression="${manifestLocation}" default-value="${project.build.outputDirectory}/META-INF"
   */
  private File m_manifestLocation;

  /**
   * The Maven project.
   * 
   * @parameter expression="${project}"
   * @required
   * @readonly
   */
  private MavenProject m_project;

  /**
   * The directory for the generated JAR.
   * 
   * @parameter expression="${project.build.directory}"
   * @required
   */
  private String m_buildDirectory;

  /**
   * Project types which this plugin supports.
   * 
   * @parameter parameter="supportedProjectTypes"
   */
  private List< String > m_supportedProjectTypes = Arrays.asList(new String[] { "deployment-package" });

  /**
   * 
   * @parameter property="localRepository" expression="${localRepository}"
   */
  private ArtifactRepository m_localRepository;

  /**
   * @parameter property="remoteRepositories" expression="${project.remoteArtifactRepositories}"
   */
  private List< ArtifactRepository > m_remoteRepositories;

  /**
   * @parameter property="writeExtraData"
   */
  private boolean m_writeExtraData = true;

  /**
   * @component
   */
  private ArtifactFactory m_artifactFactory;

  /**
   * @component
   */
  private ArtifactResolver m_artifactresolver;

  /**
   * @component
   */
  private ArtifactHandlerManager m_artifactHandlerManager;

  /**
   * @component
   */
  private ArchiverManager m_archiverManager;

  /**
   * This method will be called by the Maven framework in order to execute this plugin.
   * 
   * @throws MojoExecutionException
   *           id any error occures
   * @throws MojoFailureException
   *           id any error occures
   */
  public final void execute() throws MojoExecutionException, MojoFailureException {
    try {
      // Inject the context
      // TODO Can we do this with maven automatically?
      DeploymentPackageInfo deploymentPackageInfo = getDeploymentPackageInfo();
      if(deploymentPackageInfo == null) {
        throw new MojoExecutionException("No DeploymentPackageInfo provided");
      }
      
      for (final BundleResource bundleResource : deploymentPackageInfo.getBundleResources()) {
        bundleResource.setDeploymentContext(this);
      }

      new DeploymentPlugin(this).execute();
    } catch (final RuntimeException e) {
      throw new MojoExecutionException("Error while executing plugin ", e);
    } catch (final Throwable e) {
      throw new MojoExecutionException("Error while executing plugin ", e);
    }
  }

  /**
   * @return the logger
   * @see net.sourceforge.osgi.deployment.maven.IDeploymentPluginContext#getLogger()
   */
  public final Log getLogger() {
    return super.getLog();
  }

  /**
   * @return the outputDirectory
   * @see net.sourceforge.osgi.deployment.maven.IDeploymentPluginContext#getOutputDirectory()
   */
  public final File getOutputDirectory() {
    return m_outputDirectory;
  }

  /**
   * @param p_outputDirectory
   *          the outputDirectory to set
   */
  public final void setOutputDirectory(final File p_outputDirectory) {
    m_outputDirectory = p_outputDirectory;
  }

  /**
   * @return the base directory
   * @see net.sourceforge.osgi.deployment.maven.IDeploymentPluginContext#getBaseDir()
   */
  public final File getBaseDir() {
    return m_baseDir;
  }

  /**
   * @return the menifest location
   * @see net.sourceforge.osgi.deployment.maven.IDeploymentPluginContext#getManifestLocation()
   */
  public final File getManifestLocation() {
    return m_manifestLocation;
  }

  /**
   * @return the maven project
   * @see net.sourceforge.osgi.deployment.maven.IDeploymentPluginContext#getProject()
   */
  public final MavenProject getProject() {
    return m_project;
  }

  /**
   * @return the build directory
   * @see net.sourceforge.osgi.deployment.maven.IDeploymentPluginContext#getBuildDirectory()
   */
  public final String getBuildDirectory() {
    return m_buildDirectory;
  }

  /**
   * @return the supported packaging types
   * @see net.sourceforge.osgi.deployment.maven.IDeploymentPluginContext#getSupportedProjectTypes()
   */
  public final List< String > getSupportedProjectTypes() {
    return m_supportedProjectTypes;
  }

  /**
   * @return the local repository
   * @see net.sourceforge.osgi.deployment.maven.IDeploymentPluginContext#getLocalRepository()
   */
  public final ArtifactRepository getLocalRepository() {
    return m_localRepository;
  }

  /**
   * @return the remote repositories
   * @see net.sourceforge.osgi.deployment.maven.IDeploymentPluginContext#getRemoteRepositories()
   */
  public final List< ArtifactRepository > getRemoteRepositories() {
    return m_remoteRepositories;
  }

  /**
   * @return the artifact factory
   * @see net.sourceforge.osgi.deployment.maven.IDeploymentPluginContext#getArtifactFactory()
   */
  public final ArtifactFactory getArtifactFactory() {
    return m_artifactFactory;
  }

  /**
   * @return the artifact resolver
   * @see net.sourceforge.osgi.deployment.maven.IDeploymentPluginContext#getArtifactResolver()
   */
  public final ArtifactResolver getArtifactResolver() {
    return m_artifactresolver;
  }

  /**
   * @return the artifact handler manager
   * @see net.sourceforge.osgi.deployment.maven.IDeploymentPluginContext#getArtifactHandlerManager()
   */
  public final ArtifactHandlerManager getArtifactHandlerManager() {
    return m_artifactHandlerManager;
  }

  /**
   * @return the archiver manager
   * @see net.sourceforge.osgi.deployment.maven.IDeploymentPluginContext#getArchiverManager()
   */
  public final ArchiverManager getArchiverManager() {
    return m_archiverManager;
  }

  /**
   * @return the deployment package info
   * @see net.sourceforge.osgi.deployment.maven.IDeploymentPluginContext#getDeploymentPackageInfo()
   */
  public final DeploymentPackageInfo getDeploymentPackageInfo() {
    return m_deploymentPackageInfo;
  }

  /**
   * @return <CODE>TRUE</CODE> if extra data should be generated into the manifest file, else <CODE>FALSE></CODE>. Default is <CODE>TRUE</CODE>.
   * @see net.sourceforge.osgi.deployment.maven.IDeploymentPluginContext#isWriteExtraData()
   */
  public final boolean isWriteExtraData() {
    return m_writeExtraData;
  }

  /**
   * TODO How do we get the plugin name?
   * 
   * @return the plugin name
   * @see net.sourceforge.osgi.deployment.maven.IDeploymentPluginContext#getPluginName()
   */
  public final String getPluginName() {
    return "osgi-deployment-maven-plugin";
  }

  /**
   * TODO How do we get the plugin version?
   * 
   * @return the plugin version
   * @see net.sourceforge.osgi.deployment.maven.IDeploymentPluginContext#getPluginVersion()
   */
  public final String getPluginVersion() {
    return "0.1.2";
  }

  /**
   * @param p_baseDir
   *          the baseDir to set
   */
  public final void setBaseDir(final File p_baseDir) {
    m_baseDir = p_baseDir;
  }

  /**
   * @param p_manifestLocation
   *          the manifestLocation to set
   */
  public final void setManifestLocation(final File p_manifestLocation) {
    m_manifestLocation = p_manifestLocation;
  }

  /**
   * @param p_project
   *          the project to set
   */
  public final void setProject(final MavenProject p_project) {
    m_project = p_project;
  }

  /**
   * @param p_buildDirectory
   *          the buildDirectory to set
   */
  public final void setBuildDirectory(final String p_buildDirectory) {
    m_buildDirectory = p_buildDirectory;
  }

  /**
   * @param p_supportedProjectTypes
   *          the supportedProjectTypes to set
   */
  public final void setSupportedProjectTypes(final List< String > p_supportedProjectTypes) {
    m_supportedProjectTypes = p_supportedProjectTypes;
  }

  /**
   * @param p_localRepository
   *          the localRepository to set
   */
  public final void setLocalRepository(final ArtifactRepository p_localRepository) {
    m_localRepository = p_localRepository;
  }

  /**
   * @param p_remoteRepositories
   *          the remoteRepositories to set
   */
  public final void setRemoteRepositories(final List< ArtifactRepository > p_remoteRepositories) {
    m_remoteRepositories = p_remoteRepositories;
  }

  /**
   * @param p_artifactFactory
   *          the artifactFactory to set
   */
  public final void setArtifactFactory(final ArtifactFactory p_artifactFactory) {
    m_artifactFactory = p_artifactFactory;
  }

  /**
   * @param p_artifactResolver
   *          the artifactResolver to set
   */
  public final void setArtifactResolver(final ArtifactResolver p_artifactResolver) {
    m_artifactresolver = p_artifactResolver;
  }

  /**
   * @param p_artifactHandlerManager
   *          the artifactHandlerManager to set
   */
  public final void setArtifactHandlerManager(final ArtifactHandlerManager p_artifactHandlerManager) {
    m_artifactHandlerManager = p_artifactHandlerManager;
  }

  /**
   * @param p_archiverManager
   *          the archiverManager to set
   */
  public final void setArchiverManager(final ArchiverManager p_archiverManager) {
    m_archiverManager = p_archiverManager;
  }

  /**
   * @param p_deploymentPackageInfo
   *          the deploymentPackage to set
   */
  public final void setDeploymentPackageInfo(final DeploymentPackageInfo p_deploymentPackageInfo) {
    m_deploymentPackageInfo = p_deploymentPackageInfo;
  }

  /**
   * @param p_deploymentPackageInfo
   *          the deploymentPackage to set
   */
  public final void setDeploymentPackage(final DeploymentPackageInfo p_deploymentPackageInfo) {
    setDeploymentPackageInfo(p_deploymentPackageInfo);
  }

  /**
   * @param p_writeExtraData
   *          the writeExtraData to set
   */
  public final void setWriteExtraData(final boolean p_writeExtraData) {
    m_writeExtraData = p_writeExtraData;
  }

  /**
   * This method resolves an artifact on all available repositories and returns the file handle to that artifact.
   * 
   * @param p_groupId
   *          the groupId of the artifact to resolve
   * @param p_artifactId
   *          the artifactId of the artifact to resolve
   * @param p_version
   *          the version of the artifact to resolve
   * @return the resolved file handle of the artifact
   */
  public final File resolveResource(final String p_groupId, final String p_artifactId, final String p_version) {
    try {
      final Artifact artifact = getArtifactFactory().createArtifact(p_groupId, p_artifactId, p_version, Artifact.SCOPE_RUNTIME, "jar");
      getArtifactResolver().resolve(artifact, getRemoteRepositories(), getLocalRepository());
      final File artifactFile = artifact.getFile();
      return artifactFile;
    } catch (final Exception e) {
      // Wrap checked exception
      throw new DeploymentPluginException("Error while resolving resource " + p_groupId + ":" + p_artifactId + ":" + p_version, e);
    }
  }

}
